//KEVIN RYE - 2018
//V1.0
//ARDUINO 1.8.5



/* PINOUTS FROM EAGLE ON ATMEGA2560
  D47 PMBOT
  D46 PMTOP
  D26 BOTCOL
  D25 TOPCOL

  D24 LATCH
  D23 CLOCK
  D22 DATA

  D15 SHUTDOWN
  D14 AUDIO

  D13 AUTOSWITCH
  D12 REDSWITCH
  D11 GREENSWITCH
  D10 BLUESWITCH
  D3 RED (LED)
  D2 GREEN (LED)
  D5 BLUE (LED)

  A5 HR BUTTON
  A4 MIN BUTTON
  A3 SEC BUTTON
  A2 ALARM BUTTON
  A1 ENABLE BUTTON
  A0 MODE BUTTON
*/

#include <Wire.h>
#include <Sodaq_DS3231.h>
#include <EEPROM.h>

//RTC stuff
int seconds = 0;
int minutes = 0;
int hours = 0;
int hh = 0;
int dayOfWeek; //not used (but needed for DS3231 function), DOW is calculated and pushed to 'returnedDOWvar'
int dayOfMonth = 1;
int month = 1;
int year = 0;

//bools
bool timeMode = true;
bool dateMode = false;
bool alarmMode = false;
bool pm = false;
boolean saveNewAlarmHours = false;
boolean saveNewAlarmMinutes = false;
boolean saveNewAlarmStatus = false;

//saved alarm settings
byte savedHours;
byte savedHoursAddress = 0;
byte savedMinutes;
byte savedMinutesAddress = 1;
byte savedAlarmMode;
byte savedAlarmModeAddress = 2;

//year, month, date, hour, min, sec and week-day(starts from 0 and goes to 6)
//writing any non-existent time-data may interfere with normal operation of the RTC.
//Take care of week-day also.
DateTime dt(2018, 4, 28, 15, 39, 0, 0);

//date calc stuff
#define CalendarYrToTm(Y) ((Y) - 1970)
#define LEAP_YEAR(Y) (((1970+Y)>0) && !((1970+Y)%4) && (((1970+Y)%100) || !((1970+Y)%400)))
static  const uint8_t monthDays[] = {31, 28, 31, 30, 31, 30, 31, 31, 30, 31, 30, 31}; // API starts months from 1, this array starts from 0

#define SHIFT_DATA_PIN 22 //SERIN pin 3 on TPIC6B595
#define SHIFT_CLOCK_PIN 23 //SRCK pin 13 on TPIC6B595
#define SHIFT_LATCH_PIN 24 //RCK pin 12 on TPIC6B595

int shiftHrSection = 0;
int shiftMinSection = 0;
int shiftSecSection = 0;

//rgb
const int redPin = 3; //Red LED
const int greenPin = 2; //Green LED
const int bluePin = 5; // Blue LED
const int autoSwitch = 13;
const int redSwitch = 12;
const int greenSwitch = 11;
const int blueSwitch = 10;
unsigned int rgbColor[3];

//colons
const int top = 26;  //Col Top and Col Bottom are labelled bacwards on the PCB
const int bottom = 25; //Col Top and Col Bottom are labelled bacwards on the PCB
const int pmTop = 46;
const int pmBot = 47;

//alarm
bool setAlarmActive = false; //set up default condition, will be saved or recalled from EEPROM
byte alarmHours = 0; //set up default condition
byte adjAlarmHours;
byte alarmMinutes = 0; //set up default condition
boolean alarmIsActive = false; //when the alarm is actually sounding

//audio
byte amplifier = 14;
byte numTones = 2;
int alarmTones[] = {800, 600};
int chimeTones[] = {1200, 1200};
byte enableAmplifier = 15; //shutdown function
int speakerState = LOW;
unsigned long previousMillis = 0;
unsigned long currentMillis = 0;;
unsigned long currentMillis2 = 0;;

// constants won't change:
const long interval = 300;

//buttons
int hourMonthButton = A5;
int minDayButton = A4;
int secYearButton = A3;
int alarmButton = A2;
int enableButton = A1;
int modeButton = A0;

//shift
void nixie_driveDigit(unsigned long digit, bool final_digit);
void nixie_shiftDigit(unsigned long binLong, bool latch_high);

boolean toggle = true;
boolean tubesaver = false;
int counter = 0;

void setup() {

  pinMode(SHIFT_LATCH_PIN, OUTPUT);
  pinMode(SHIFT_DATA_PIN, OUTPUT);
  pinMode(SHIFT_CLOCK_PIN, OUTPUT);

  pinMode(redPin, OUTPUT);
  pinMode(greenPin, OUTPUT);
  pinMode(bluePin, OUTPUT);
  pinMode(autoSwitch, INPUT);
  pinMode(redSwitch, INPUT);
  pinMode(greenSwitch, INPUT);
  pinMode(blueSwitch, INPUT);

  pinMode(top, OUTPUT);
  pinMode(bottom, OUTPUT);
  pinMode(pmTop, OUTPUT);
  pinMode(pmBot, OUTPUT);
  pinMode(top, OUTPUT);
  pinMode(bottom, OUTPUT);

  pinMode(hourMonthButton, INPUT);
  pinMode(minDayButton, INPUT);
  pinMode(secYearButton, INPUT);
  pinMode(alarmButton, INPUT);
  pinMode(enableButton, INPUT);
  pinMode(modeButton, INPUT);

  pinMode(enableAmplifier, OUTPUT);
  pinMode(amplifier, OUTPUT);

  //set initial states
  setColorRgb(0, 0, 0);
  digitalWrite(top, LOW);
  digitalWrite(bottom, LOW);
  digitalWrite(pmTop, LOW);
  digitalWrite(pmBot, LOW);
  digitalWrite(top, LOW);
  digitalWrite(bottom, LOW);
  digitalWrite(enableAmplifier, LOW);

  //pull in alarm time from EEPROM
  //read saved hours for alarm from EEPROM
  savedHours = EEPROM.read(savedHoursAddress);
  if (savedHours != 0) {
    alarmHours = savedHours;
  }
  else {
    alarmHours = 0;
  }

  //read saved minutes for alarm from EEPROM
  savedMinutes = EEPROM.read(savedMinutesAddress);
  if (savedMinutes != 0) {
    alarmMinutes = savedMinutes;
  }
  else {
    alarmMinutes = 0;
  }

  //read saved alarm mode from EEPROM
  savedAlarmMode = EEPROM.read(savedAlarmModeAddress);
  if (savedAlarmMode == 0) { //0 = NO //1 = YES
    setAlarmActive = false;
  }
  else {
    setAlarmActive = true;
  }

  Wire.begin();
  rtc.begin();
  //Serial.begin(9600);
  // rtc.setDateTime(dt); //Adjust date-time as defined 'dt' above

}

void loop() {

  //RGB
  if (digitalRead(redSwitch) == LOW) {

    digitalWrite(redPin, HIGH);
    digitalWrite(greenPin, LOW);
    digitalWrite(bluePin, LOW);
  }

  if (digitalRead(greenSwitch) == LOW) {

    digitalWrite(redPin, LOW);
    digitalWrite(greenPin, HIGH);
    digitalWrite(bluePin, LOW);
  }

  if (digitalRead(blueSwitch) == LOW) {

    digitalWrite(redPin, LOW);
    digitalWrite(greenPin, LOW);
    digitalWrite(bluePin, HIGH);
  }

  if (digitalRead(autoSwitch) == LOW) {

    // Start off with red.
    rgbColor[0] = 255;
    rgbColor[1] = 0;
    rgbColor[2] = 0;

    // Choose the colors to increment and decrement.
    for (int decColor = 0; decColor < 3; decColor += 1) {
      int incColor = decColor == 2 ? 0 : decColor + 1;

      // cross-fade the two colors.
      for (int i = 0; i < 255; i += 1) {
        rgbColor[decColor] -= 1;
        rgbColor[incColor] += 1;

        setColorRgb(rgbColor[0], rgbColor[1], rgbColor[2]);
        delay(1); //remove delay in final sketch
      }
    }
  }

  //poll the DS3231 for the date and time
  get_time();
  get_date();

  //pass RTC values into a new variable -
  //this way one variable stores time in 24 hour mode and another that can be converted to 12 hour mode
  //time is monitored 'in the background' in 24 hour mode using the RTC, but is displayed in an adjusted 12-hr format
  hh = hours;

  //convert to 12 hr mode for display
  if (hours < 1) {
    hh = 12;
  }

  if ((hours > 12) && (hours < 24)) {
    hh = hours - 12;
  }

  //pm ind
  if ((hours > 11) && (true == timeMode)) {
    pm = true;
  }
  else {
    pm = false;
  }


  //handle button presses
  //TIME MODE
  //hour/month button pressed
  if ((digitalRead(hourMonthButton) == LOW) && (true == timeMode)) {

    //Serial.println("hour");
    //debounce
    delay(200);

    hours++;

    if (hours > 23) {
      hours = 0;
    }

    set_time();
  }

  //min/day button pressed
  if ((digitalRead(minDayButton) == LOW) && (true == timeMode)) {
    //Serial.println("min");
    //debounce
    delay(200);

    minutes++;
    seconds = 0;

    if (minutes > 59) {
      minutes = 0;
    }

    set_time();
  }

  //sec/year button pressed
  if ((digitalRead(secYearButton) == LOW) && (true == timeMode)) {
    //Serial.println("sec");
    //debounce
    delay(200);

    seconds = 0;
    set_time();
  }

  //mode button pressed
  if ((digitalRead(modeButton) == LOW) && (true == timeMode) && (false == tubesaver)) {
    //Serial.println("mode");
    //debounce
    delay(200);

    timeMode = false;
    alarmMode = false;
    dateMode = true;
  }

  if ((digitalRead(modeButton) == LOW) && (true == dateMode) && (false == tubesaver)) {
    //Serial.println("mode");
    //debounce
    delay(200);

    timeMode = true;
    alarmMode = false;
    dateMode = false;
  }

  if ((digitalRead(modeButton) == LOW) && (true == alarmMode) && (false == tubesaver)) {
    //Serial.println("mode");
    //debounce
    delay(200);

    timeMode = false;
    alarmMode = false;
    dateMode = true;
  }

  //cancel time saver mode
  if ((digitalRead(modeButton) == LOW) && (true == tubesaver)) {

    //debounce
    delay(200);

    tubesaver = false;
    timeMode = true;
    alarmMode = false;
    dateMode = false;
  }

  //in alarmMode
  if (true == alarmMode) {

    //adjust alarm time for 12 hours
    adjAlarmHours = alarmHours;

    if (adjAlarmHours > 12) {
      adjAlarmHours = alarmHours - 12;
    }
    if (adjAlarmHours == 0) {
      adjAlarmHours = 12;
    }

    //enable am/pm nixie as appropriate
    if (alarmHours > 11) {
      pm = true;
    }
    else {
      pm = false;
    }
  }

  //ALARM MODE
  //hour/month button pressed
  if ((digitalRead(hourMonthButton) == LOW) && (true == alarmMode)) {

    //debounce
    delay(200);

    alarmHours++;

    if (alarmHours > 23) {
      alarmHours = 0;
    }

    if (alarmHours != savedHours ) {

      saveNewAlarmHours = true;
      writeToEEPROM();
    }
  }

  //min/day button pressed
  if ((digitalRead(minDayButton) == LOW) && (true == alarmMode)) {

    //debounce
    delay(200);

    alarmMinutes++;

    if (alarmMinutes > 59) {
      alarmMinutes = 0;
    }

    if (alarmMinutes != savedMinutes ) {

      saveNewAlarmMinutes = true;
      writeToEEPROM();
    }
  }

  if ((digitalRead(secYearButton) == LOW) && (true == alarmMode)) {

    //does nothing, seconds remain at 0 in alarm mode
  }

  //enable button pressed
  if (digitalRead(enableButton) == LOW) {

    //debounce
    delay(200);

    //toggle alarm status
    setAlarmActive = !setAlarmActive;

    //save new setting
    saveNewAlarmStatus = true;
    writeToEEPROM();
  }

  //alarm menu button pressed
  if ((digitalRead(alarmButton) == LOW) && (true == alarmMode)) {

    //debaounce
    delay(200);

    timeMode = true;
    alarmMode = false;
    dateMode = false;
  }

  if ((digitalRead(alarmButton) == LOW) && (false == alarmMode)) {

    //debaounce
    delay(200);

    timeMode = false;
    alarmMode = true;
    dateMode = false;
  }

  //DATE MODE
  //hour/month button pressed
  if ((digitalRead(hourMonthButton) == LOW) && (true == dateMode) && (false == alarmMode)) {

    //debounce
    delay(200);

    //set month
    month++;
    if (month > 12) month = 1;

    //send update to RTC
    set_date();
  }

  //min/day button pressed
  if ((digitalRead(minDayButton) == LOW) && (true == dateMode) && (false == alarmMode)) {

    //debounce
    delay(200);

    //set day
    dayOfMonth++;

    //if feb
    if (month == 2) {
      if (dayOfMonth > 28) dayOfMonth = 1;
    }

    //if leap year
    //is it worth it?

    //if month has 30 days: Apr, Jun, Sep, Nov
    if ((month == 4) || (month == 6) || (month == 9) || (month == 11))  {
      if (dayOfMonth > 30) dayOfMonth = 1;
    }

    //if month has 31 days: Jan, Mar, May, Jul, Aug, Oct, Dec
    if ((month == 1) || (month == 3) || (month == 5) || (month == 7) || (month == 8) || (month == 10)) {
      if (dayOfMonth > 31) dayOfMonth = 1;
    }

    //send update to RTC
    set_date();
  }

  //year button pressed
  if ((digitalRead(secYearButton) == LOW) && (true == dateMode) && (false == alarmMode)) {

    //debounce
    delay(200);

    year++;
    if (year > 99) year = 0;

    //send update to RTC
    set_date();
  }

  //chime on hour
  //chime only when in time mode and when not alarming
  if ((minutes == 0) && (seconds < 1) && (true == timeMode) && (false == alarmIsActive) ) {

    digitalWrite(enableAmplifier, HIGH);

    //this gives me 2 beeps
    unsigned long currentMillis2 = millis();

    if (currentMillis2 - previousMillis >= interval) {
      // save the last time you pulsed the speaker
      previousMillis = currentMillis2;

      // if the speaker is off turn it on and vice-versa:
      if (speakerState == LOW) {
        speakerState = HIGH;
        for (int i = 0; i < numTones; i++)
        {
          tone(amplifier, alarmTones[i]);
          //Serial.println("chime tone");
        }
        delay(200); //putting in a delay here kills me, but the loop is too long for this to be heard otherwise. Total hack, I know.
      }

      else {
        speakerState = LOW;
        noTone(amplifier);
        digitalWrite(enableAmplifier, LOW);
      }
    }
  }

  if ((adjAlarmHours == hh) && (alarmMinutes == minutes) && (seconds < 20)) {

    if ((true == setAlarmActive) && (true == timeMode) ) {
      digitalWrite(enableAmplifier, HIGH);

      //this gives me 3 beeps
      unsigned long currentMillis2 = millis();

      alarmIsActive = true;

      if (currentMillis2 - previousMillis >= interval) {
        // save the last time you pulsed the speaker
        previousMillis = currentMillis2;

        // if the speaker is off turn it on and vice-versa:
        if (speakerState == LOW) {
          speakerState = HIGH;
          for (int i = 0; i < numTones; i++)
          {
            tone(amplifier, alarmTones[i]);
            //Serial.println("chime tone");
          }
          delay(200); //putting in a delay here kills me, but the loop is too long for this to be heard otherwise. Total hack, I know.
        }

        else {
          speakerState = LOW;
          alarmIsActive = false;
          noTone(amplifier);
          digitalWrite(enableAmplifier, LOW);
        }
      }
    }
  }

  //enable tube saver
  //settings seconds to 10 so that this does not trigger when ssetting the time.
  //(Seconds are always reset to 0 when settings the time so this can never be true when setting the clock.)
  if ((hours == 0) && (minutes == 0) && (seconds == 10)) {
    tubesaver = true;
  }

  if ((hours == 6) && (minutes == 0) && (seconds == 10)) {
    tubesaver = false;
  }

  //finally, push data to nixies
  displayData();

} // end loop


// Displays the time
void displayData() {

  //shift out time
  if ((true == timeMode) && (false == tubesaver)) {
    shiftHrSection = hh;
    shiftMinSection = minutes;
    shiftSecSection = seconds;

    //colons
    digitalWrite(top, HIGH);
    digitalWrite(bottom, HIGH);

    //pm
    if (true == pm) {
      digitalWrite(pmTop, HIGH);

    } else {
      digitalWrite(pmTop, LOW);
    }
  } //end shift out time

  else if ((true == alarmMode) && (false == tubesaver)) { //shift out alaarm
    shiftHrSection = adjAlarmHours;
    shiftMinSection = alarmMinutes;
    shiftSecSection = 0;

    //colons
    digitalWrite(top, LOW);
    digitalWrite(bottom, LOW);

    //pm
    if (true == pm) {
      digitalWrite(pmTop, HIGH);

    } else {
      digitalWrite(pmTop, LOW);
    }

    //alarm enabled ind
    if (true == setAlarmActive) {
      digitalWrite(pmBot, HIGH);

    } else {
      digitalWrite(pmBot, LOW);
    }
  } //end shift out alarm mode

  else if ((true == dateMode) && (false == tubesaver)) { //shift out date (dateMode = true)
    shiftHrSection = month;
    shiftMinSection = dayOfMonth;
    shiftSecSection = year;

    //colons
    digitalWrite(top, LOW);
    digitalWrite(bottom, HIGH);

    //pm
    if (true == pm) {
      digitalWrite(pmTop, LOW);
    }
  }

  //tube saver mode - count from 12am to 6am
  else if (true == tubesaver) {

    counter = counter + 11;
    delay(100);

    if (counter > 99) {
      counter = 0;
    }

    shiftHrSection = counter;
    shiftMinSection = counter;
    shiftSecSection = counter;

    //colons OFF
    digitalWrite(top, LOW);
    digitalWrite(bottom, LOW);

    //pm
    //no action - PM is off anyway

    //alarm
    digitalWrite(pmBot, LOW);
  } //end shift out tube saver

  //shift out data
  nixie_driveDigit(shiftMinSection, false);
  nixie_driveDigit(shiftSecSection, true);
  nixie_driveDigit(shiftHrSection, false);

}

//nixie shift code provided with the help of Chad Etzel. Many thanks!
// using nixie_driveDigit() you can pass in any two digit number from 0 - 99.
// this will convert the decimal into the binary representation to shift out.
void nixie_driveDigit(unsigned long digit, bool final_digit) {
  //Serial.println(digit, DEC);

 // if (digit > 59) {
   // nixie_shiftDigit(0);
 // } else {

    unsigned long lsd = digit % 10;
    unsigned long msd = (digit / 10) % 10;
    //Serial.println(lsd, DEC);
    //Serial.println(msd, DEC);
    unsigned long binLong = ((unsigned long)((unsigned long)1 << (((unsigned long)10 + msd)))) | ((unsigned long)((unsigned long)1 << (unsigned long)lsd));

    nixie_shiftDigit(binLong, final_digit);
  // }
}

void nixie_shiftDigit(unsigned long binLong, bool latch_high) {
  //Serial.println(binLong, BIN);

  digitalWrite(SHIFT_LATCH_PIN, LOW);
  char data = binLong & 0xff;
  shiftOut(SHIFT_DATA_PIN, SHIFT_CLOCK_PIN, LSBFIRST, data);


  data = (binLong >> 8) & 0xff;
  shiftOut(SHIFT_DATA_PIN, SHIFT_CLOCK_PIN, LSBFIRST, data);

  data = (binLong >> 16) & 0xff;
  shiftOut(SHIFT_DATA_PIN, SHIFT_CLOCK_PIN, LSBFIRST, data);

  if (latch_high) {
    digitalWrite(SHIFT_LATCH_PIN, HIGH);
  }
}

void setColorRgb(unsigned int red, unsigned int green, unsigned int blue) {
  analogWrite(redPin, red);
  analogWrite(greenPin, green);
  analogWrite(bluePin, blue);
}

//save settings to EEPROM
void writeToEEPROM() {

  //only save the settings that were changed

  //save alarm hours if it was changed
  if (true == saveNewAlarmHours) {
    EEPROM.write(savedHoursAddress, alarmHours);
  }

  //save alarm minutes if it was changed
  if (true == saveNewAlarmMinutes) {
    EEPROM.write(savedMinutesAddress, alarmMinutes);
  }

  //save alarm status if it was changed
  if (true == saveNewAlarmStatus) {
    EEPROM.write(savedAlarmModeAddress, savedAlarmMode);
  }

  //set flags back to false, so subsequent changes will also be saved
  saveNewAlarmHours = false;
  saveNewAlarmMinutes = false;
  saveNewAlarmStatus = false;
}

//DS3231 stuff
void set_date() {
  Wire.beginTransmission(104);
  Wire.write(3);//set register to 3 (day)
  Wire.write(decToBcd(dayOfWeek)); //this is actually useless. Could be anything really. The day of week is being calculated above based on the date, not what's stored on the DS3231
  Wire.write(decToBcd(dayOfMonth));
  Wire.write(decToBcd(month));
  Wire.write(decToBcd(year));
  Wire.endTransmission();
  //Serial.println("set date");
}

void get_date() {
  Wire.beginTransmission(104);
  Wire.write(3);//set register to 3 (day)
  Wire.endTransmission();
  Wire.requestFrom(104, 4); //get 5 bytes(day,date,month,year,control);
  dayOfWeek   = bcdToDec(Wire.read()); //this is actually useless. Could be anything really. The day of week is being calculated above based on the date, not what's stored on the DS3231
  dayOfMonth  = bcdToDec(Wire.read());
  month = bcdToDec(Wire.read());
  year  = bcdToDec(Wire.read());
}

void set_time() {
  Wire.beginTransmission(104);
  Wire.write(0);
  Wire.write(decToBcd(seconds));
  Wire.write(decToBcd(minutes));
  Wire.write(decToBcd(hours));
  Wire.endTransmission();
  //Serial.println("set time");
}

void get_time() {
  Wire.beginTransmission(104);
  Wire.write(0);//set register to 0
  Wire.endTransmission();
  Wire.requestFrom(104, 3);//get 3 bytes (seconds,minutes,hours);
  seconds = bcdToDec(Wire.read() & 0x7f);
  minutes = bcdToDec(Wire.read());
  hours = bcdToDec(Wire.read() & 0x3f);
}

byte decToBcd(byte val) {
  return ((val / 10 * 16) + (val % 10));
}

byte bcdToDec(byte val) {
  return ((val / 16 * 10) + (val % 16));
}


//calculate how many days are in the displayed month
uint8_t getMonthDays(uint8_t y, uint8_t m) {
  return ((m == 2) && LEAP_YEAR(y)) ? 29 : monthDays[m - 1];
}

//calculate the day of week based on the given date
byte calcDayOfWeek(unsigned int y, byte m, byte d)
{
  // cast out multiples of 400 from y
  y -= (2000 * (y >> 11));
  while (y >= 400) y -= 400;

  boolean leap = ((y & 3) == 0);
  if ((y == 100) || (y == 200) || (y == 300)) leap = false;

  if (d > 31 || d == 0) return 0;

  byte w = 6;
  while (y >= 100) {
    y -= 100;
    w -= 2;
  }

  w += (y + (y >> 2));

  // correction for Jan. and Feb. of leap year
  if (leap && (m <= 2)) w--;

  // using subtraction iso addition makes the while at end possible 1 iteration faster.
  switch (m)
  {
    case 1:
      w++;
      break;
    case 2:
      if (d > (leap ? 29 : 28)) return 0;
      w += 4;
      break;
    case 3:
      w += 4;
      break;
    case 5:
      w += 2;
      break;
    case 7:
      break;
    case 8:
      w += 3;
      break;
    case 10:
      w++;
      break;
    case 12:
      w += 6;
      break;
    default:
      if (d > 30) return 0;
      switch (m)
      {
        case 4:
          break;
        case 6:
          w += 5;
          break;
        case 9:
          w += 6;
          break;
        case 11:
          w += 4;
          break;
        default:
          return 0;
      }
  }

  w += d;

  // there are only 7 days in a week, so we "cast out" sevens
  while (w > 7) w = (w >> 3) + (w & 7);
  return w;
}

